{****************************************************}
{  Reservoir simulation unit. Copyright (c) 2007 by  }
{  National Technical University of Athens           }
{  Authors, Stefanos Kozanis, Andreas Efstratiadis.  }
{****************************************************}

unit reservoir;

interface

uses Classes, Contnrs, interpol, ts, dates;

type
  TRunoffType = (trtHeight, trtVolume);

type
{** A point structure for coefficient curves of target objects.
}
  TCoefPoint = record
    Level, Coefficient: Real;
  end;

type
{** A class for target objects.
    @Author Stefanos
}
  TTarget = class(TPersistent)
  private
    FPriority: Integer;
    FCurve: array of TCoefPoint;
    FTimeseries: TTimeseries;
    FSupplySeries: TTimeseries;
    FCurveFilename, FTimeseriesFilename, FSupplyFilename: string;
    FIsDownstream: Boolean;
  public
    constructor Create;
    destructor Destroy; override;
    procedure LoadCurveFromFile;
    procedure LoadTimeseriesFromfile;
    procedure WriteSupplyTimeseries; overload;
    procedure WriteSupplyTimeseries(AOutputDirectory: string); overload;
    function Interpolate(Stage: Real): Real;
    property Priority: Integer read FPriority write FPriority;
    property DemandTimeseries: TTimeseries read FTimeseries write
      FTimeseries;
    property ActualSupplyTimeseries: TTimeseries read FSupplySeries write
      FSupplySeries;
    property IsDownstream: Boolean read FIsDownStream write FIsDownStream;
    property CurveFilename: string read FCurveFilename write FCurveFilename;
    property TimeseriesFilename: string read FTimeseriesFilename
      write FTimeseriesFilename;
    property ActualSupplyTimeseriesFilename: string read FSupplyFilename write
      FSupplyFilename;
  end;

{** A class for target sets (object lists).
    @Author Stefanos
}
  TTargetSet = class(TPersistent)
  private
    FObjectList: TObjectList;
    FOutputDirectory: string;
    function GetCount: Integer;
  protected
    function Get(Index: Integer): TTarget;
    procedure Put(Index: Integer; Item: TTarget);
  public
    constructor Create;
    destructor Destroy; override;
    procedure Clear;
    function First: TTarget;
    function Last: TTarget;
    procedure Delete(Index: Integer);
    property Items[Index: Integer]: TTarget read Get write Put; default;
    property OutputDirectory: string read FOutputDirectory
      write FOutputDirectory;
    property Count: Integer read GetCount;
    procedure Add(Priority: Integer; IsDownstream: Boolean;
      CurveFilename, TimeseriesFilename, SupplyFilename: string);
    procedure LoadAll;
    procedure SaveAll;
    function CalcOutflow(ADate: TDateTime; AStageVolumeCurve:TInterpolatingCurve;
      MinVolume: Real; var CurrentVolume: Real): Real;
  end;

type
{** A class for reservoir objects.
    @Author Stefanos
}
  TReservoir = class(TPersistent)
  private
    FStageVolumeCurve: TInterpolatingCurve;
    FStageAreaCurve: TInterpolatingCurve;
    FStageLeakageCurve: TLeakageInterpolation;
    procedure AssessVolumes;
    procedure Initialize;
  public
    LowerStage, IntakeLevel, SpillLevel, BasinArea: Real;
    MinVolume, DeadVolume, MaxVolume: Real;
    DivertSpill: Boolean;
    constructor Create;
    destructor Destroy; override;
    procedure LoadCurves(StageAreaVolumeFile, StageLeakageFile: string);
    property StageVolumeCurve: TInterpolatingCurve read FStageVolumeCurve
      write FStageVolumeCurve;
    property StageAreaCurve: TInterpolatingCurve read FStageAreaCurve
      write FStageAreaCurve;
    property StageLeakageCurve: TLeakageInterpolation read
      FStageLeakageCurve write FStageLeakageCurve;
  end;

function NextStep(ADate: TDateTime; ATimestep: TTimeStep): TDateTime;

{** A function for executing one step reservoir simulation.
    @Author Stefanos
}
function CalcReservoirNewStage(AReservoir: TReservoir; ATargetSet: TTargetSet;
  ADate: TDateTime;
  CurrentStage, RainfallHeight, EvaporationHeight, Runoff,
  MiscInflowVolume, MiscOutflowVolume: Real;
  RunoffType: TRunoffType;
  var SpillVolume, CurrentVolume, NewVolume, Outflow: Real): Real;

{** A function for executing one step reservoir simulation within the
    SRM enviroment.
    @Author Stefanos
}
procedure RunSRM(AReservoir: TReservoir; ATargetSet: TTargetSet;
  PreviousStage, InputRunoff: Real;
  CurDate: TDateTime;
  Rainfall, Evaporation, MiscInflowVolume,
  MiscOutflowVolume: TTimeseries; RunoffType: TRunoffType;
  var CurrentVolume, NewStage, NewVolume, Outflow, Spill: Real);

implementation

uses SysUtils, istrutils, DateUtils, Math, tsprocess;

{ Local functions }

procedure LoadStageAreaVolumeCurves(AStageVolumeCurve, AStageAreaCurve:
  TInterpolatingCurve; AStageAreaVolumeFile: string);
var
  ACurvePoint: TCurvePoint;
  s: string;
  ADelimiter, ADecimalSeparator, ASavedDecimalSeparator: Char;
  F: TextFile;
  AStage, AVolume, AArea: Real;
begin
  ADelimiter := ',';
  ADecimalSeparator := '.';
  AssignFile(F, AStageAreaVolumeFile);
  ASavedDecimalSeparator := SysUtils.DecimalSeparator;
  try
    SysUtils.DecimalSeparator := ADecimalSeparator;
    Reset(F);
    AStageVolumeCurve.Clear;
    AStageAreaCurve.Clear;
    while not Eof(F) do
    begin
      ReadLn(F,s);
      AStage := StrToFloat(DelimitedStringItem(s, 1, ADelimiter));
      AVolume := StrToFloat(DelimitedStringItem(s, 2, ADelimiter));
      AArea := StrToFloat(DelimitedStringItem(s, 3, ADelimiter));
      if AArea <=0 then AArea := 0.00001;
      if AVolume <=0 then AVolume := 0.00001;
      ACurvePoint.Independent := AStage;
      ACurvePoint.Dependent := AArea;
      AStageAreaCurve.Add(ACurvePoint);
      ACurvePoint.Independent := AStage;
      ACurvePoint.Dependent := AVolume;
      AStageVolumeCurve.Add(ACurvePoint);
    end;
  finally
    SysUtils.DecimalSeparator := ASavedDecimalSeparator;
    CloseFile(F);
  end;
end;

procedure LoadStageLeakageCurve(AStageLeakageCurve: TLeakageInterpolation;
  AStageLeakageFile: string);
var
  s: string;
  ADelimiter, ADecimalSeparator, ASavedDecimalSeparator: Char;
  F: TextFile;
  a,b,c,e: Real;
  i: Integer;
begin
  ADelimiter := ',';
  ADecimalSeparator := '.';
  AssignFile(F, AStageLeakageFile);
  ASavedDecimalSeparator := SysUtils.DecimalSeparator;
  try
    SysUtils.DecimalSeparator := ADecimalSeparator;
    Reset(F);
    for i := 1 to 12 do
    begin
      ReadLn(F,s);
      a := StrToFloat(DelimitedStringItem(s, 1, ADelimiter));
      b := StrToFloat(DelimitedStringItem(s, 2, ADelimiter));
      c := StrToFloat(DelimitedStringItem(s, 3, ADelimiter));
      e := StrToFloat(DelimitedStringItem(s, 4, ADelimiter));
      AStageLeakageCurve.FACoef[i] := a;
      AStageLeakageCurve.FBCoef[i] := b;
      AStageLeakageCurve.FCCoef[i] := c;
      AStageLeakageCurve.FECoef[i] := e;
    end;
  finally
    SysUtils.DecimalSeparator := ASavedDecimalSeparator;
    CloseFile(F);
  end;
end;

{ TTarget }

constructor TTarget.Create;
begin
  inherited Create;
  FTimeseries := nil;
  FSupplySeries := nil;
  FTimeseries := TTimeseries.Create;
  FSupplySeries := TTimeseries.Create;
  FIsDownstream := False;
  FTimeseriesFilename := '';
  FSupplyFilename := '';
  FCurveFilename := '';
  SetLength(FCurve, 0);
  FPriority := 0;
end;

destructor TTarget.Destroy;
begin
  FTimeseries.Free;
  FSupplySeries.Free;
  SetLength(FCurve, 0);
  inherited Destroy;
end;

function TTarget.Interpolate(Stage: Real): Real;
var
  i: Integer;
begin
  Result := 0;
  for i := 0 to Length(FCurve)-1 do
    if Stage>=FCurve[i].Level then
      Result := FCurve[i].Coefficient;
end;

procedure TTarget.LoadCurveFromFile;
var
  F: TextFile;
  ACoefPoint: TCoefPoint;
  s: string;
  ADelimiter, ADecimalSeparator, ASavedDecimalSeparator: Char;
begin
  ADelimiter := ',';
  ADecimalSeparator := '.';
  AssignFile(F, CurveFilename);
  ASavedDecimalSeparator := SysUtils.DecimalSeparator;
  try
    SysUtils.DecimalSeparator := ADecimalSeparator;
    Reset(F);
    SetLength(FCurve, 0);
    while not Eof(F) do
    begin
      Readln(F, s);
      ACoefPoint.Level := StrToFloat(DelimitedStringItem(s, 1, ADelimiter));
      ACoefPoint.Coefficient :=
        StrToFloat(DelimitedStringItem(s, 2, ADelimiter));
      SetLength(FCurve, Length(FCurve)+1);
      FCurve[Length(FCurve)-1] := ACoefPoint;
    end;
  finally
    SysUtils.DecimalSeparator := ASavedDecimalSeparator;
    Close(F);
  end;
end;

procedure TTarget.LoadTimeseriesFromfile;
begin
  if not FileExists(FTimeseriesFilename) then Exit;
  FTimeseries.LoadFromFile(FTimeseriesFilename);
end;

procedure TTarget.WriteSupplyTimeseries;
begin
  WriteSupplyTimeseries('');
end;

procedure TTarget.WriteSupplyTimeseries(AOutputDirectory: string);
begin
  if FSupplyFilename = '' then Exit;
  FSupplySeries.WriteToFile(AOutputDirectory+FSupplyFilename);
end;

function CompareProc(Item1, Item2: Pointer): Integer;
begin
  Result := TTarget(Item1).Priority - TTarget(Item2).Priority;
end;

{ TTargetSet }

constructor TTargetSet.Create;
begin
  inherited Create;
  FObjectList := TObjectList.Create(True);
  FOutputDirectory := '';
end;

destructor TTargetSet.Destroy;
begin
  FObjectList.Free;
  inherited Destroy;
end;

function TTargetSet.GetCount: Integer;
begin
  Result := FObjectList.Count;
end;

resourcestring
  rsIndexOutOfBounds = 'Index out of bounds';

function TTargetSet.Get(Index: Integer): TTarget;
begin
  if (Index<0) or (Index>=Count) then
    raise EListError.Create(rsIndexOutOfBounds);
  Result := FObjectList.Items[Index] as TTarget;
end;

procedure TTargetSet.Put(Index: Integer; Item: TTarget);
begin
  if (Index<0) or (Index>=Count) then
    raise EListError.Create(rsIndexOutOfBounds);
  FObjectList.Items[Index] := Item;
end;

procedure TTargetSet.Clear;
begin
  FObjectList.Clear;
end;

function TTargetSet.First: TTarget;
begin
  Result := FObjectList.First as TTarget;
end;

function TTargetSet.Last: TTarget;
begin
  Result := FObjectList.Last as TTarget;
end;

procedure TTargetSet.Delete(Index: Integer);
begin
  FObjectList.Delete(Index);
end;

procedure TTargetSet.Add(Priority: Integer; IsDownstream: Boolean;
  CurveFilename, TimeseriesFilename, SupplyFilename: string);
var
  ATarget: TTarget;
begin
  ATarget := nil;
  try
    ATarget := TTarget.Create;
    ATarget.Priority := Priority;
    ATarget.CurveFilename := CurveFilename;
    ATarget.TimeseriesFilename := TimeseriesFilename;
    ATarget.ActualSupplyTimeseriesFilename := SupplyFilename;
    ATarget.IsDownstream := IsDownstream;
    ATarget.ActualSupplyTimeseries.Clear;
    FObjectList.Add(ATarget);
    ATarget := nil;
    FObjectList.Sort(CompareProc);
  finally
    ATarget.Free;
  end;
end;

procedure TTargetSet.LoadAll;
var
  i: Integer;
begin
  for i := 0 to Count-1 do
    with Items[i] do
    begin
      if FileExists(TimeseriesFilename) then LoadTimeseriesFromfile;
      if FileExists(CurveFilename) then LoadCurveFromFile;
    end;
end;

procedure TTargetSet.SaveAll;
var
  i: Integer;
begin
  for i := 0 to Count-1 do
    with Items[i] do
      WriteSupplyTimeseries(FOutputDirectory);
end;

function TTargetSet.CalcOutflow(ADate: TDateTime; AStageVolumeCurve:TInterpolatingCurve;
  MinVolume: Real; var CurrentVolume: Real): Real;
var
  i: Integer;
  CurrentStage: Real;
  AOutflow: Real;

  function GetSingleOutflow(ATarget: TTarget): Real;
  var
    AIndex: Integer;
  begin
    Result := 0;
    with ATarget do
    begin
      AIndex := DemandTimeseries.IndexOf(ADate);
      if AIndex<0 then Exit;
      Result := Max(0,
        Min(CurrentVolume - MinVolume, Interpolate(CurrentStage)*
          DemandTimeseries[AIndex].AsFloat));
    end;
  end;

  procedure PutOutflowToTimeseries(ATarget: TTarget; AOutflow: Real);
  var
    AIndex: Integer;
  begin
    with ATarget do
      with ActualSupplyTimeseries do
      begin
        AIndex := ActualSupplyTimeseries.IndexOf(ADate);
        if AIndex>0 then ActualSupplyTimeseries.Delete(AIndex);
        ActualSupplyTimeseries.Add(ADate, False, AOutflow, '', msNew);
      end;
  end;

begin
  Result := 0;
  CurrentStage := AStageVolumeCurve.ReverseInterpolate(CurrentVolume);
  for i := 0 to Count-1 do
  begin
    AOutflow := GetSingleOutflow(Items[i]);
    PutOutflowToTimeseries(Items[i], AOutflow);
    if Items[i].IsDownstream then Result := Result + AOutflow;
    CurrentVolume := CurrentVolume - AOutflow;
    CurrentStage := AStageVolumeCurve.ReverseInterpolate(CurrentVolume);
  end;
end;

{ TReservoir }

constructor TReservoir.Create;
begin
  inherited Create;
  FStageVolumeCurve := nil;
  FStageAreaCurve := nil;
  FStageLeakageCurve := nil;
  DivertSpill := True;
  try
    FStageVolumeCurve := TInterpolatingCurve.Create(True);
    FStageAreaCurve := TInterpolatingCurve.Create(True);
    FStageLeakageCurve := TLeakageInterpolation.Create;
  except
    FStageVolumeCurve.Free;
    FStageAreaCurve.Free;
    FStageLeakageCurve.Free;
    raise;
  end;
end;

destructor TReservoir.Destroy;
begin
  FStageVolumeCurve.Free;
  FStageAreaCurve.Free;
  FStageLeakageCurve.Free;
  inherited Destroy;
end;

procedure TReservoir.Initialize;
var
  i: Integer;
begin
  FStageVolumeCurve.Clear;
  FStageAreaCurve.Clear;
  for i := 1 to 12 do
    with FStageLeakageCurve do
    begin
      FACoef[i] := 0;
      FBCoef[i] := 0;
      FCCoef[i] := 0;
      FECoef[i] := 0;
    end;
end;

procedure TReservoir.LoadCurves(StageAreaVolumeFile, StageLeakageFile: string);
begin
  if not FileExists(StageAreaVolumeFile) then
    raise Exception.Create('Stage- Area, Volume file should be supplied');
  Initialize;
  LoadStageAreaVolumeCurves(FStageVolumeCurve, FStageAreaCurve,
    StageAreaVolumeFile);
  if FileExists(StageLeakageFile) then
    LoadStageLeakageCurve(FStageLeakageCurve, StageLeakageFile);
  AssessVolumes;
end;

procedure TReservoir.AssessVolumes;
begin
  MinVolume := 0.00001;
  DeadVolume := FStageVolumeCurve.Interpolate(IntakeLevel);
  MaxVolume := FStageVolumeCurve.Interpolate(SpillLevel);
end;

{ Exported functions }

function NextStep(ADate: TDateTime; ATimestep: TTimeStep): TDateTime;
begin
  Result := -1; //dummy
  if ATimestep = tstDaily then
    Result := AddDateTime(ADate, 1)
  else if ATimeStep = tstMonthly then
    Result := IncMonth(ADate)
  else
    Assert(False, 'Unsupported time step');
end;

function CalcReservoirNewStage(AReservoir: TReservoir; ATargetSet: TTargetSet;
  ADate: TDateTime;
  CurrentStage, RainfallHeight, EvaporationHeight, Runoff,
  MiscInflowVolume, MiscOutflowVolume: Real;
  RunoffType: TRunoffType;  
  var SpillVolume, CurrentVolume, NewVolume, Outflow: Real): Real;
var
  AArea: Real;
  ALeakage: Real;
  ARunoffTerm: Real;
begin
  SpillVolume := 0;
  with AReservoir do
  begin
    ALeakage := StageLeakageCurve.Interpolate(MonthOf(ADate), CurrentStage);
    CurrentVolume := StageVolumeCurve.Interpolate(CurrentStage);
    AArea := StageAreaCurve.Interpolate(CurrentStage);
{Water Balance}
    ARunoffTerm := Runoff;
    if RunoffType = trtHeight then
      ARunoffTerm := ARunoffTerm * (BasinArea - AArea) * 0.001
    else
      ARunoffTerm := ARunoffTerm * (BasinArea - AArea)/BasinArea;
    NewVolume := CurrentVolume + RainfallHeight * AArea * 0.001 +
      ARunoffTerm +
      MiscInflowVolume -
      EvaporationHeight * AArea * 0.001 - ALeakage - MiscOutflowVolume;
    Outflow := ATargetSet.CalcOutflow(ADate, StageVolumeCurve,
      AReservoir.DeadVolume, NewVolume);
{Level range check, results}
    if NewVolume<MinVolume then NewVolume := MinVolume;
    SpillVolume := Max(0, NewVolume - MaxVolume);
    if NewVolume>MaxVolume then NewVolume := MaxVolume;
    Result := StageVolumeCurve.ReverseInterpolate(NewVolume);
    if Result < LowerStage then Result := LowerStage;
    if Result > SpillLevel then Result := SpillLevel;
  end;
end;

procedure RunSRM(AReservoir: TReservoir; ATargetSet: TTargetSet;
  PreviousStage, InputRunoff: Real;
  CurDate: TDateTime;
  Rainfall, Evaporation, MiscInflowVolume,
  MiscOutflowVolume: TTimeseries; RunoffType: TRunoffType;
  var CurrentVolume, NewStage, NewVolume, Outflow, Spill: Real);
var
  ACurrentStage, ANewStage, ARainfall, AEvaporation, ARunoff,
    AInflows, AOutflows: Real;
  ASpill, ACurrentVolume, ANewVolume: Real;
  ADate: TDateTime;

  function GetTimeseriesValue(ATimeseries: TTimeseries; ADate: TDateTime): Real;
  var
    AIndex: Integer;
  begin
    AIndex := ATimeseries.IndexOf(ADate);
    Result := 0;
    if AIndex>-1 then
      if not ATimeseries[AIndex].IsNull then
        Result := ATimeseries[AIndex].AsFloat;
  end;

  procedure LoadValues(ADate: TDateTime);
  begin
    ARainfall := GetTimeseriesValue(Rainfall, ADate);
    AEvaporation := GetTimeseriesValue(Evaporation, ADate);
    ARainfall := GetTimeseriesValue(Rainfall, ADate);
    AInflows := GetTimeseriesValue(MiscInflowVolume, ADate);
    AOutflows := GetTimeseriesValue(MiscOutflowVolume, ADate);
  end;

begin
  ARunoff := InputRunoff;
  ADate := CurDate;
  ACurrentStage := PreviousStage;
  LoadValues(ADate);
  ANewStage := CalcReservoirNewStage(AReservoir, ATargetSet, ADate, ACurrentStage,
    ARainfall, AEvaporation, ARunoff, AInflows, AOutflows,
    RunoffType, ASpill, ACurrentVolume, ANewVolume, Outflow);
  Spill := ASpill;
  CurrentVolume := ACurrentVolume;
  NewVolume := ANewVolume;
  if AReservoir.DivertSpill then
    Outflow := Outflow+ASpill;
  NewStage := ANewStage;
end;

end.
